"""Token Manager classes.

There should be a 1-to-1 mapping between an instance of a subclass of
:class:`.BaseTokenManager` and a :class:`.Reddit` instance.

A few proof of concept token manager classes are provided here, but it is expected that
PRAW users will create their own token manager classes suitable for their needs.

.. deprecated:: 7.4.0

    Tokens managers have been deprecated and will be removed in the near future.

"""

from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING

from . import _deprecate_args

if TYPE_CHECKING:  # pragma: no cover
    import prawcore

    import praw


class BaseTokenManager(ABC):
    """An abstract class for all token managers."""

    @abstractmethod
    def post_refresh_callback(self, authorizer: prawcore.auth.BaseAuthorizer):
        """Handle callback that is invoked after a refresh token is used.

        :param authorizer: The ``prawcore.Authorizer`` instance used containing
            ``access_token`` and ``refresh_token`` attributes.

        This function will be called after refreshing the access and refresh tokens.
        This callback can be used for saving the updated ``refresh_token``.

        """

    @abstractmethod
    def pre_refresh_callback(self, authorizer: prawcore.auth.BaseAuthorizer):
        """Handle callback that is invoked before refreshing PRAW's authorization.

        :param authorizer: The ``prawcore.Authorizer`` instance used containing
            ``access_token`` and ``refresh_token`` attributes.

        This callback can be used to inspect and modify the attributes of the
        ``prawcore.Authorizer`` instance, such as setting the ``refresh_token``.

        """

    @property
    def reddit(self) -> praw.Reddit:
        """Return the :class:`.Reddit` instance bound to the token manager."""
        return self._reddit

    @reddit.setter
    def reddit(self, value: praw.Reddit):
        if self._reddit is not None:
            msg = "'reddit' can only be set once and is done automatically"
            raise RuntimeError(msg)
        self._reddit = value

    def __init__(self):
        """Initialize a :class:`.BaseTokenManager` instance."""
        self._reddit = None


class FileTokenManager(BaseTokenManager):
    """Provides a single-file based token manager.

    It is expected that the file with the initial ``refresh_token`` is created prior to
    use.

    .. warning::

        The same ``file`` should not be used by more than one instance of this class
        concurrently. Doing so may result in data corruption. Consider using
        :class:`.SQLiteTokenManager` if you want more than one instance of PRAW to
        concurrently manage a specific ``refresh_token`` chain.

    """

    def __init__(self, filename: str):
        """Initialize a :class:`.FileTokenManager` instance.

        :param filename: The file the contains the refresh token.

        """
        super().__init__()
        self._filename = filename

    def post_refresh_callback(self, authorizer: prawcore.auth.BaseAuthorizer):
        """Update the saved copy of the refresh token."""
        with Path(self._filename).open("w") as fp:
            fp.write(authorizer.refresh_token)

    def pre_refresh_callback(self, authorizer: prawcore.auth.BaseAuthorizer):
        """Load the refresh token from the file."""
        if authorizer.refresh_token is None:
            with Path(self._filename).open() as fp:
                authorizer.refresh_token = fp.read().strip()


class SQLiteTokenManager(BaseTokenManager):
    """Provides a SQLite3 based token manager.

    Unlike, :class:`.FileTokenManager`, the initial database need not be created ahead
    of time, as it'll automatically be created on first use. However, initial refresh
    tokens will need to be registered via :meth:`.register` prior to use.

    .. warning::

        This class is untested on Windows because we encountered file locking issues in
        the test environment.

    """

    @_deprecate_args("database", "key")
    def __init__(self, *, database: str, key: str):
        """Initialize a :class:`.SQLiteTokenManager` instance.

        :param database: The path to the SQLite database.
        :param key: The key used to locate the refresh token. This ``key`` can be
            anything. You might use the ``client_id`` if you expect to have unique a
            refresh token for each ``client_id``, or you might use a redditor's
            ``username`` if you're managing multiple users' authentications.

        """
        super().__init__()
        import sqlite3

        self._connection = sqlite3.connect(database)
        self._connection.execute(
            "CREATE TABLE IF NOT EXISTS tokens (id, refresh_token, updated_at)"
        )
        self._connection.execute(
            "CREATE UNIQUE INDEX IF NOT EXISTS ux_tokens_id on tokens(id)"
        )
        self._connection.commit()
        self.key = key

    def _get(self):
        cursor = self._connection.execute(
            "SELECT refresh_token FROM tokens WHERE id=?", (self.key,)
        )
        result = cursor.fetchone()
        if result is None:
            raise KeyError
        return result[0]

    def _set(self, refresh_token: str):
        """Set the refresh token in the database.

        This function will overwrite an existing value if the corresponding ``key``
        already exists.

        """
        self._connection.execute(
            "REPLACE INTO tokens VALUES (?, ?, datetime('now'))",
            (self.key, refresh_token),
        )
        self._connection.commit()

    def is_registered(self) -> bool:
        """Return whether ``key`` already has a ``refresh_token``."""
        cursor = self._connection.execute(
            "SELECT refresh_token FROM tokens WHERE id=?", (self.key,)
        )
        return cursor.fetchone() is not None

    def post_refresh_callback(self, authorizer: prawcore.auth.BaseAuthorizer):
        """Update the refresh token in the database."""
        self._set(authorizer.refresh_token)

        # While the following line is not strictly necessary, it ensures that the
        # refresh token is not used elsewhere. And also forces the pre_refresh_callback
        # to always load the latest refresh_token from the database.
        authorizer.refresh_token = None

    def pre_refresh_callback(self, authorizer: prawcore.auth.BaseAuthorizer):
        """Load the refresh token from the database."""
        assert authorizer.refresh_token is None
        authorizer.refresh_token = self._get()

    def register(self, refresh_token: str) -> bool:
        """Register the initial refresh token in the database.

        :returns: ``True`` if ``refresh_token`` is saved to the database, otherwise,
            ``False`` if there is already a ``refresh_token`` for the associated
            ``key``.

        """
        cursor = self._connection.execute(
            "INSERT OR IGNORE INTO tokens VALUES (?, ?, datetime('now'))",
            (self.key, refresh_token),
        )
        self._connection.commit()
        return cursor.rowcount == 1



"""Provide the RedditBase class."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any
from urllib.parse import urlparse

from ...endpoints import API_PATH
from ...exceptions import InvalidURL
from ..base import PRAWBase

if TYPE_CHECKING:  # pragma: no cover
    import praw


class RedditBase(PRAWBase):
    """Base class that represents actual Reddit objects."""

    @staticmethod
    def _url_parts(url: str) -> list[str]:
        parsed = urlparse(url)
        if not parsed.netloc:
            raise InvalidURL(url)
        return parsed.path.rstrip("/").split("/")

    def __eq__(self, other: Any | str) -> bool:
        """Return whether the other instance equals the current."""
        if isinstance(other, str):
            return other.lower() == str(self).lower()
        return (
            isinstance(other, self.__class__)
            and str(self).lower() == str(other).lower()
        )

    def __getattr__(self, attribute: str) -> Any:
        """Return the value of ``attribute``."""
        if not attribute.startswith("_") and not self._fetched:
            self._fetch()
            return getattr(self, attribute)
        msg = f"{self.__class__.__name__!r} object has no attribute {attribute!r}"
        raise AttributeError(msg)

    def __hash__(self) -> int:
        """Return the hash of the current instance."""
        return hash(self.__class__.__name__) ^ hash(str(self).lower())

    def __init__(
        self,
        reddit: praw.Reddit,
        _data: dict[str, Any] | None,
        _extra_attribute_to_check: str | None = None,
        _fetched: bool = False,
        _str_field: bool = True,
    ):
        """Initialize a :class:`.RedditBase` instance.

        :param reddit: An instance of :class:`.Reddit`.

        """
        super().__init__(reddit, _data=_data)
        self._fetched = _fetched
        if _str_field and self.STR_FIELD not in self.__dict__:
            if (
                _extra_attribute_to_check is not None
                and _extra_attribute_to_check in self.__dict__
            ):
                return
            msg = f"An invalid value was specified for {self.STR_FIELD}. Check that the argument for the {self.STR_FIELD} parameter is not empty."
            raise ValueError(msg)

    def __ne__(self, other: object) -> bool:
        """Return whether the other instance differs from the current."""
        return not self == other

    def __repr__(self) -> str:
        """Return an object initialization representation of the instance."""
        return f"{self.__class__.__name__}({self.STR_FIELD}={str(self)!r})"

    def __str__(self) -> str:
        """Return a string representation of the instance."""
        return getattr(self, self.STR_FIELD)

    def _fetch(self):
        self._fetched = True

    def _fetch_data(self):
        name, fields, params = self._fetch_info()
        path = API_PATH[name].format(**fields)
        return self._reddit.request(method="GET", params=params, path=path)

    def _reset_attributes(self, *attributes: str):
        for attribute in attributes:
            if attribute in self.__dict__:
                del self.__dict__[attribute]
        self._fetched = False



"""Provide the InboxToggleableMixin class."""

from ....const import API_PATH


class InboxToggleableMixin:
    """Interface for classes that can optionally receive inbox replies."""

    def disable_inbox_replies(self):
        """Disable inbox replies for the item.

        .. note::

            This can only apply to items created by the authenticated user.

        Example usage:

        .. code-block:: python

            comment = reddit.comment("dkk4qjd")
            comment.disable_inbox_replies()

            submission = reddit.submission("8dmv8z")
            submission.disable_inbox_replies()

        .. seealso::

            :meth:`.enable_inbox_replies`

        """
        self._reddit.post(
            API_PATH["sendreplies"], data={"id": self.fullname, "state": False}
        )

    def enable_inbox_replies(self):
        """Enable inbox replies for the item.

        .. note::

            This can only apply to items created by the authenticated user.

        Example usage:

        .. code-block:: python

            comment = reddit.comment("dkk4qjd")
            comment.enable_inbox_replies()

            submission = reddit.submission("8dmv8z")
            submission.enable_inbox_replies()

        .. seealso::

            :meth:`.disable_inbox_replies`

        """
        self._reddit.post(
            API_PATH["sendreplies"], data={"id": self.fullname, "state": True}
        )



"""Contains functions dealing with snake case conversions."""

from __future__ import annotations

import re
from typing import Any

_re_camel_to_snake = re.compile(r"([a-z0-9](?=[A-Z])|[A-Z](?=[A-Z][a-z]))")


def camel_to_snake(name: str) -> str:
    """Convert ``name`` from camelCase to snake_case."""
    return _re_camel_to_snake.sub(r"\1_", name).lower()


def snake_case_keys(dictionary: dict[str, Any]) -> dict[str, Any]:
    """Return a new dictionary with keys converted to snake_case.

    :param dictionary: The dict to be corrected.

    """
    return {camel_to_snake(k): v for k, v in dictionary.items()}



"""Provide the RedditorList class."""

from .base import BaseList


class RedditorList(BaseList):
    """A list of :class:`.Redditor` objects. Works just like a regular list."""

    CHILD_ATTRIBUTE = "children"



"""Provide the PRAWBase superclass."""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:  # pragma: no cover
    import praw


class PRAWBase:
    """Superclass for all models in PRAW."""

    @staticmethod
    def _safely_add_arguments(
        *, arguments: dict[str, Any], key: str, **new_arguments: Any
    ):
        """Replace arguments[key] with a deepcopy and update.

        This method is often called when new parameters need to be added to a request.
        By calling this method and adding the new or updated parameters we can insure we
        don't modify the dictionary passed in by the caller.

        """
        value = deepcopy(arguments[key]) if key in arguments else {}
        value.update(new_arguments)
        arguments[key] = value

    @classmethod
    def parse(cls, data: dict[str, Any], reddit: praw.Reddit) -> Any:
        """Return an instance of ``cls`` from ``data``.

        :param data: The structured data.
        :param reddit: An instance of :class:`.Reddit`.

        """
        return cls(reddit, _data=data)

    def __init__(self, reddit: praw.Reddit, _data: dict[str, Any] | None):
        """Initialize a :class:`.PRAWBase` instance.

        :param reddit: An instance of :class:`.Reddit`.

        """
        self._reddit = reddit
        if _data:
            for attribute, value in _data.items():
                setattr(self, attribute, value)



"""Provide the Stylesheet class."""

from .base import PRAWBase


class Stylesheet(PRAWBase):
    """Represent a stylesheet.

    .. include:: ../../typical_attributes.rst

    ============== ========================================
    Attribute      Description
    ============== ========================================
    ``images``     A list of images used by the stylesheet.
    ``stylesheet`` The contents of the stylesheet, as CSS.
    ============== ========================================

    """



"""Provide the InboxableMixin class."""

from ....const import API_PATH


class InboxableMixin:
    """Interface for :class:`.RedditBase` subclasses that originate from the inbox."""

    def block(self):
        """Block the user who sent the item.

        .. note::

            This method pertains only to objects which were retrieved via the inbox.

        Example usage:

        .. code-block:: python

            comment = reddit.comment("dkk4qjd")
            comment.block()

            # or, identically:

            comment.author.block()

        """
        self._reddit.post(API_PATH["block"], data={"id": self.fullname})

    def collapse(self):
        """Mark the item as collapsed.

        .. note::

            This method pertains only to objects which were retrieved via the inbox.

        Example usage:

        .. code-block:: python

            inbox = reddit.inbox()

            # select first inbox item and collapse it message = next(inbox)
            message.collapse()

        .. seealso::

            :meth:`.uncollapse`

        """
        self._reddit.inbox.collapse([self])

    def mark_read(self):
        """Mark a single inbox item as read.

        .. note::

            This method pertains only to objects which were retrieved via the inbox.

        Example usage:

        .. code-block:: python

            inbox = reddit.inbox.unread()

            for message in inbox:
                # process unread messages
                ...

        .. seealso::

            :meth:`.mark_unread`

        To mark the whole inbox as read with a single network request, use
        :meth:`.Inbox.mark_all_read`

        """
        self._reddit.inbox.mark_read([self])

    def mark_unread(self):
        """Mark the item as unread.

        .. note::

            This method pertains only to objects which were retrieved via the inbox.

        Example usage:

        .. code-block:: python

            inbox = reddit.inbox(limit=10)

            for message in inbox:
                # process messages
                ...

        .. seealso::

            :meth:`.mark_read`

        """
        self._reddit.inbox.mark_unread([self])

    def unblock_subreddit(self):
        """Unblock a subreddit.

        .. note::

            This method pertains only to objects which were retrieved via the inbox.

        For example, to unblock all blocked subreddits that you can find by going
        through your inbox:

        .. code-block:: python

            from praw.models import SubredditMessage

            subs = set()
            for item in reddit.inbox.messages(limit=None):
                if isinstance(item, SubredditMessage):
                    if (
                        item.subject == "[message from blocked subreddit]"
                        and str(item.subreddit) not in subs
                    ):
                        item.unblock_subreddit()
                        subs.add(str(item.subreddit))

        """
        self._reddit.post(API_PATH["unblock_subreddit"], data={"id": self.fullname})

    def uncollapse(self):
        """Mark the item as uncollapsed.

        .. note::

            This method pertains only to objects which were retrieved via the inbox.

        Example usage:

        .. code-block:: python

            inbox = reddit.inbox()

            # select first inbox item and uncollapse it
            message = next(inbox)
            message.uncollapse()

        .. seealso::

            :meth:`.collapse`

        """
        self._reddit.inbox.uncollapse([self])



"""Provide the GildableMixin class."""

from warnings import warn

from ....const import API_PATH
from ....util import _deprecate_args


class GildableMixin:
    """Interface for classes that can be gilded."""

    @_deprecate_args("gild_type", "is_anonymous", "message")
    def award(
        self,
        *,
        gild_type: str = "gid_2",
        is_anonymous: bool = True,
        message: str = None,
    ) -> dict:
        """Award the author of the item.

        :param gild_type: Type of award to give. See table below for currently know
            global award types.
        :param is_anonymous: If ``True``, the authenticated user's username will not be
            revealed to the recipient.
        :param message: Message to include with the award.

        :returns: A dict containing info similar to what is shown below:

            .. code-block:: python

                {
                    "subreddit_balance": 85260,
                    "treatment_tags": [],
                    "coins": 8760,
                    "gildings": {"gid_1": 0, "gid_2": 1, "gid_3": 0},
                    "awarder_karma_received": 4,
                    "all_awardings": [
                        {
                            "giver_coin_reward": 0,
                            "subreddit_id": None,
                            "is_new": False,
                            "days_of_drip_extension": 0,
                            "coin_price": 75,
                            "id": "award_9663243a-e77f-44cf-abc6-850ead2cd18d",
                            "penny_donate": 0,
                            "coin_reward": 0,
                            "icon_url": "https://www.redditstatic.com/gold/awards/icon/SnooClappingPremium_512.png",
                            "days_of_premium": 0,
                            "icon_height": 512,
                            "tiers_by_required_awardings": None,
                            "icon_width": 512,
                            "static_icon_width": 512,
                            "start_date": None,
                            "is_enabled": True,
                            "awardings_required_to_grant_benefits": None,
                            "description": "For an especially amazing showing.",
                            "end_date": None,
                            "subreddit_coin_reward": 0,
                            "count": 1,
                            "static_icon_height": 512,
                            "name": "Bravo Grande!",
                            "icon_format": "APNG",
                            "award_sub_type": "PREMIUM",
                            "penny_price": 0,
                            "award_type": "global",
                            "static_icon_url": "https://i.redd.it/award_images/t5_q0gj4/59e02tmkl4451_BravoGrande-Static.png",
                        }
                    ],
                }


        .. warning::

            Requires the authenticated user to own Reddit Coins. Calling this method
            will consume Reddit Coins.

        To award the gold award anonymously do:

        .. code-block:: python

            comment = reddit.comment("dkk4qjd")
            comment.award()

            submission = reddit.submission("8dmv8z")
            submission.award()

        To award the platinum award with the message 'Nice!' and reveal your username to
        the recipient do:

        .. code-block:: python

            comment = reddit.comment("dkk4qjd")
            comment.award(gild_type="gild_3", message="Nice!", is_anonymous=False)

            submission = reddit.submission("8dmv8z")
            submission.award(gild_type="gild_3", message="Nice!", is_anonymous=False)

        .. include:: awards.txt

        """
        params = {
            "api_type": "json",
            "gild_type": gild_type,
            "is_anonymous": is_anonymous,
            "thing_id": self.fullname,
            "message": message,
        }
        return self._reddit.post(API_PATH["award_thing"], params=params)

    def gild(self) -> dict:
        """Alias for :meth:`.award` to maintain backwards compatibility."""
        warn(
            "'.gild' has been renamed to '.award'.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.award()



"""List of API endpoints PRAW knows about."""

# fmt: off
API_PATH = {
    "about_edited":            "r/{subreddit}/about/edited/",
    "about_log":               "r/{subreddit}/about/log/",
    "about_modqueue":          "r/{subreddit}/about/modqueue/",
    "about_reports":           "r/{subreddit}/about/reports/",
    "about_spam":              "r/{subreddit}/about/spam/",
    "about_sticky":            "r/{subreddit}/about/sticky/",
    "about_stylesheet":        "r/{subreddit}/about/stylesheet/",
    "about_traffic":           "r/{subreddit}/about/traffic/",
    "about_unmoderated":       "r/{subreddit}/about/unmoderated/",
    "accept_mod_invite":       "r/{subreddit}/api/accept_moderator_invite",
    "add_subreddit_rule":      "api/add_subreddit_rule",
    "add_whitelisted":         "api/add_whitelisted",
    "approve":                 "api/approve/",
    "award_thing":             "api/v2/gold/gild",
    "block":                   "api/block",
    "block_user":              "/api/block_user/",
    "blocked":                 "prefs/blocked/",
    "collapse":                "api/collapse_message/",
    "collection":              "api/v1/collections/collection",
    "collection_add_post":     "api/v1/collections/add_post_to_collection",
    "collection_create":       "api/v1/collections/create_collection",
    "collection_delete":       "api/v1/collections/delete_collection",
    "collection_desc":         "api/v1/collections/update_collection_description",
    "collection_layout":       "api/v1/collections/update_collection_display_layout",
    "collection_follow":       "api/v1/collections/follow_collection",
    "collection_remove_post":  "api/v1/collections/remove_post_in_collection",
    "collection_reorder":      "api/v1/collections/reorder_collection",
    "collection_subreddit":    "api/v1/collections/subreddit_collections",
    "collection_title":        "api/v1/collections/update_collection_title",
    "comment":                 "api/comment/",
    "comment_replies":         "message/comments/",
    "compose":                 "api/compose/",
    "contest_mode":            "api/set_contest_mode/",
    "convert_rte_body":        "api/convert_rte_body_format",
    "del":                     "api/del/",
    "delete_message":          "api/del_msg",
    "delete_sr_banner":        "r/{subreddit}/api/delete_sr_banner",
    "delete_sr_header":        "r/{subreddit}/api/delete_sr_header",
    "delete_sr_icon":          "r/{subreddit}/api/delete_sr_icon",
    "delete_sr_image":         "r/{subreddit}/api/delete_sr_img",
    "deleteflair":             "r/{subreddit}/api/deleteflair",
    "distinguish":             "api/distinguish/",
    "draft":                   "api/v1/draft",
    "drafts":                  "api/v1/drafts",
    "domain":                  "domain/{domain}/",
    "duplicates":              "duplicates/{submission_id}/",
    "edit":                    "api/editusertext/",
    "emoji_delete":            "api/v1/{subreddit}/emoji/{emoji_name}",
    "emoji_lease":             "api/v1/{subreddit}/emoji_asset_upload_s3.json",
    "emoji_list":              "api/v1/{subreddit}/emojis/all",
    "emoji_update":            "api/v1/{subreddit}/emoji_permissions",
    "emoji_upload":            "api/v1/{subreddit}/emoji.json",
    "flair":                   "r/{subreddit}/api/flair/",
    "flairconfig":             "r/{subreddit}/api/flairconfig/",
    "flaircsv":                "r/{subreddit}/api/flaircsv/",
    "flairlist":               "r/{subreddit}/api/flairlist/",
    "flairselector":           "r/{subreddit}/api/flairselector/",
    "flairtemplate_v2":        "r/{subreddit}/api/flairtemplate_v2",
    "flairtemplateclear":      "r/{subreddit}/api/clearflairtemplates/",
    "flairtemplatedelete":     "r/{subreddit}/api/deleteflairtemplate/",
    "flairtemplatereorder":    "r/{subreddit}/api/flair_template_order",
    "friend":                  "r/{subreddit}/api/friend/",
    "friend_v1":               "api/v1/me/friends/{user}",
    "friends":                 "api/v1/me/friends/",
    "gild_user":               "api/v1/gold/give/{username}/",
    "hide":                    "api/hide/",
    "ignore_reports":          "api/ignore_reports/",
    "inbox":                   "message/inbox/",
    "info":                    "api/info/",
    "karma":                   "api/v1/me/karma",
    "leavecontributor":        "api/leavecontributor",
    "link_flair":              "r/{subreddit}/api/link_flair_v2",
    "list_banned":             "r/{subreddit}/about/banned/",
    "list_contributor":        "r/{subreddit}/about/contributors/",
    "list_moderator":          "r/{subreddit}/about/moderators/",
    "list_invited_moderator":  "/api/v1/{subreddit}/moderators_invited",
    "list_muted":              "r/{subreddit}/about/muted/",
    "list_wikibanned":         "r/{subreddit}/about/wikibanned/",
    "list_wikicontributor":    "r/{subreddit}/about/wikicontributors/",
    "live_accept_invite":      "api/live/{id}/accept_contributor_invite",
    "live_add_update":         "api/live/{id}/update",
    "live_close":              "api/live/{id}/close_thread",
    "live_contributors":       "live/{id}/contributors",
    "live_discussions":        "live/{id}/discussions",
    "live_focus":              "live/{thread_id}/updates/{update_id}",
    "live_info":               "api/live/by_id/{ids}",
    "live_invite":             "api/live/{id}/invite_contributor",
    "live_leave":              "api/live/{id}/leave_contributor",
    "live_now":                "api/live/happening_now",
    "live_remove_contrib":     "api/live/{id}/rm_contributor",
    "live_remove_invite":      "api/live/{id}/rm_contributor_invite",
    "live_remove_update":      "api/live/{id}/delete_update",
    "live_report":             "api/live/{id}/report",
    "live_strike":             "api/live/{id}/strike_update",
    "live_update_perms":       "api/live/{id}/set_contributor_permissions",
    "live_update_thread":      "api/live/{id}/edit",
    "live_updates":            "live/{id}",
    "liveabout":               "api/live/{id}/about/",
    "livecreate":              "api/live/create",
    "lock":                    "api/lock/",
    "marknsfw":                "api/marknsfw/",
    "me":                      "api/v1/me",
    "media_asset":             "api/media/asset.json",
    "mentions":                "message/mentions",
    "message":                 "message/messages/{id}/",
    "messages":                "message/messages/",
    "mod_notes":               "api/mod/notes",
    "mod_notes_bulk":          "api/mod/notes/recent",
    "moderated":               "user/{user}/moderated_subreddits/",
    "moderator_messages":      "r/{subreddit}/message/moderator/",
    "moderator_unread":        "r/{subreddit}/message/moderator/unread/",
    "modmail_archive":         "api/mod/conversations/{id}/archive",
    "modmail_bulk_read":       "api/mod/conversations/bulk/read",
    "modmail_conversation":    "api/mod/conversations/{id}",
    "modmail_conversations":   "api/mod/conversations/",
    "modmail_highlight":       "api/mod/conversations/{id}/highlight",
    "modmail_mute":            "api/mod/conversations/{id}/mute",
    "modmail_read":            "api/mod/conversations/read",
    "modmail_subreddits":      "api/mod/conversations/subreddits",
    "modmail_unarchive":       "api/mod/conversations/{id}/unarchive",
    "modmail_unmute":          "api/mod/conversations/{id}/unmute",
    "modmail_unread":          "api/mod/conversations/unread",
    "modmail_unread_count":    "api/mod/conversations/unread/count",
    "morechildren":            "api/morechildren/",
    "multireddit":             "user/{user}/m/{multi}/",
    "multireddit_api":         "api/multi/user/{user}/m/{multi}/",
    "multireddit_base":        "api/multi/",
    "multireddit_copy":        "api/multi/copy/",
    "multireddit_rename":      "api/multi/rename/",
    "multireddit_update":      "api/multi/user/{user}/m/{multi}/r/{subreddit}",
    "multireddit_user":        "api/multi/user/{user}/",
    "mute_sender":             "api/mute_message_author/",
    "my_contributor":          "subreddits/mine/contributor/",
    "my_moderator":            "subreddits/mine/moderator/",
    "my_multireddits":         "api/multi/mine/",
    "my_subreddits":           "subreddits/mine/subscriber/",
    "post_requirements":       "api/v1/{subreddit}/post_requirements",
    "preferences":             "api/v1/me/prefs",
    "quarantine_opt_in":       "api/quarantine_optin",
    "quarantine_opt_out":      "api/quarantine_optout",
    "read_all_messages":       "api/read_all_messages",
    "read_message":            "api/read_message/",
    "removal_comment_message": "api/v1/modactions/removal_comment_message",
    "removal_link_message":    "api/v1/modactions/removal_link_message",
    "removal_reasons":         "api/v1/modactions/removal_reasons",
    "removal_reason":          "api/v1/{subreddit}/removal_reasons/{id}",
    "removal_reasons_list":    "api/v1/{subreddit}/removal_reasons",
    "remove_subreddit_rule":   "api/remove_subreddit_rule",
    "remove_whitelisted":      "api/remove_whitelisted",
    "remove":                  "api/remove/",
    "reorder_subreddit_rules": "api/reorder_subreddit_rules",
    "report":                  "api/report/",
    "rules":                   "r/{subreddit}/about/rules",
    "save":                    "api/save/",
    "search":                  "r/{subreddit}/search/",
    "select_flair":            "r/{subreddit}/api/selectflair/",
    "sendreplies":             "api/sendreplies",
    "sent":                    "message/sent/",
    "set_original_content":    "api/set_original_content",
    "setpermissions":          "r/{subreddit}/api/setpermissions/",
    "show_comment":            "api/show_comment",
    "site_admin":              "api/site_admin/",
    "spoiler":                 "api/spoiler/",
    "sticky_submission":       "api/set_subreddit_sticky/",
    "store_visits":            "api/store_visits",
    "structured_styles":       "api/v1/structured_styles/{subreddit}",
    "style_asset_lease":       "api/v1/style_asset_upload_s3/{subreddit}",
    "sub_recommended":         "api/recommend/sr/{subreddits}",
    "submission":              "comments/{id}/",
    "submission_replies":      "message/selfreply/",
    "submit":                  "api/submit/",
    "submit_gallery_post":     "api/submit_gallery_post.json",
    "submit_poll_post":        "api/submit_poll_post",
    "subreddit":               "r/{subreddit}/",
    "subreddit_about":         "r/{subreddit}/about/",
    "subreddit_filter":        "api/filter/user/{user}/f/{special}/r/{subreddit}",
    "subreddit_filter_list":   "api/filter/user/{user}/f/{special}",
    "subreddit_random":        "r/{subreddit}/random/",
    "subreddit_settings":      "r/{subreddit}/about/edit/",
    "subreddit_stylesheet":    "r/{subreddit}/api/subreddit_stylesheet/",
    "subreddits_by_topic":     "api/subreddits_by_topic",
    "subreddits_default":      "subreddits/default/",
    "subreddits_gold":         "subreddits/premium/",
    "subreddits_name_search":  "api/search_reddit_names/",
    "subreddits_new":          "subreddits/new/",
    "subreddits_popular":      "subreddits/popular/",
    "subreddits_search":       "subreddits/search/",
    "subscribe":               "api/subscribe/",
    "suggested_sort":          "api/set_suggested_sort/",
    "trophies":                "api/v1/user/{user}/trophies",
    "trusted":                 "prefs/trusted",
    "unblock_subreddit":       "api/unblock_subreddit",
    "uncollapse":              "api/uncollapse_message/",
    "unfriend":                "r/{subreddit}/api/unfriend/",
    "unhide":                  "api/unhide/",
    "unignore_reports":        "api/unignore_reports/",
    "unlock":                  "api/unlock/",
    "unmarknsfw":              "api/unmarknsfw/",
    "unmute_sender":           "api/unmute_message_author/",
    "unread":                  "message/unread/",
    "unread_message":          "api/unread_message/",
    "unsave":                  "api/unsave/",
    "unspoiler":               "api/unspoiler/",
    "update_crowd_control":    "api/update_crowd_control_level",
    "update_settings":         "api/v1/subreddit/update_settings",
    "update_subreddit_rule":   "api/update_subreddit_rule",
    "upload_image":            "r/{subreddit}/api/upload_sr_img",
    "user":                    "user/{user}/",
    "user_about":              "user/{user}/about/",
    "user_by_fullname":        "/api/user_data_by_account_ids",
    "user_flair":              "r/{subreddit}/api/user_flair_v2",
    "username_available":      "api/username_available",
    "users_new":               "users/new",
    "users_popular":           "users/popular",
    "users_search":            "users/search",
    "vote":                    "api/vote/",
    "widget_create":           "r/{subreddit}/api/widget",
    "widget_lease":            "r/{subreddit}/api/widget_image_upload_s3",
    "widget_modify":           "r/{subreddit}/api/widget/{widget_id}",
    "widget_order":            "r/{subreddit}/api/widget_order/{section}",
    "widgets":                 "r/{subreddit}/api/widgets",
    "wiki_discussions":        "r/{subreddit}/wiki/discussions/{page}",
    "wiki_edit":               "r/{subreddit}/api/wiki/edit",
    "wiki_page":               "r/{subreddit}/wiki/{page}",
    "wiki_page_editor":        "r/{subreddit}/api/wiki/alloweditor/{method}",
    "wiki_page_revisions":     "r/{subreddit}/wiki/revisions/{page}",
    "wiki_page_settings":      "r/{subreddit}/wiki/settings/{page}",
    "wiki_pages":              "r/{subreddit}/wiki/pages/",
    "wiki_revert":             "r/{subreddit}/api/wiki/revert",
    "wiki_revisions":          "r/{subreddit}/wiki/revisions/",
}



"""Provides the User class."""

from __future__ import annotations

from typing import TYPE_CHECKING, Iterator
from warnings import warn

from prawcore import Conflict

from ..const import API_PATH
from ..exceptions import ReadOnlyException
from ..models import Preferences
from ..util import _deprecate_args
from ..util.cache import cachedproperty
from .base import PRAWBase
from .listing.generator import ListingGenerator
from .reddit.redditor import Redditor
from .reddit.subreddit import Subreddit

if TYPE_CHECKING:  # pragma: no cover
    import praw.models


class User(PRAWBase):
    """The :class:`.User` class provides methods for the currently authenticated user."""

    @cachedproperty
    def preferences(self) -> praw.models.Preferences:
        """Get an instance of :class:`.Preferences`.

        The preferences can be accessed as a ``dict`` like so:

        .. code-block:: python

            preferences = reddit.user.preferences()
            print(preferences["show_link_flair"])

        Preferences can be updated via:

        .. code-block:: python

            reddit.user.preferences.update(show_link_flair=True)

        The :meth:`.Preferences.update` method returns the new state of the preferences
        as a ``dict``, which can be used to check whether a change went through. Changes
        with invalid types or parameter names fail silently.

        .. code-block:: python

            original_preferences = reddit.user.preferences()
            new_preferences = reddit.user.preferences.update(invalid_param=123)
            print(original_preferences == new_preferences)  # True, no change

        """
        return Preferences(self._reddit)

    def __init__(self, reddit: praw.Reddit):
        """Initialize an :class:`.User` instance.

        This class is intended to be interfaced with through ``reddit.user``.

        """
        super().__init__(reddit, _data=None)

    def blocked(self) -> list[praw.models.Redditor]:
        r"""Return a :class:`.RedditorList` of blocked :class:`.Redditor`\ s."""
        return self._reddit.get(API_PATH["blocked"])

    def contributor_subreddits(
        self, **generator_kwargs: str | int | dict[str, str]
    ) -> Iterator[praw.models.Subreddit]:
        r"""Return a :class:`.ListingGenerator` of contributor :class:`.Subreddit`\ s.

        These are subreddits in which the user is an approved user.

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        To print a list of the subreddits that you are an approved user in, try:

        .. code-block:: python

            for subreddit in reddit.user.contributor_subreddits(limit=None):
                print(str(subreddit))

        """
        return ListingGenerator(
            self._reddit, API_PATH["my_contributor"], **generator_kwargs
        )

    @_deprecate_args("user")
    def friends(
        self, *, user: str | praw.models.Redditor | None = None
    ) -> list[praw.models.Redditor] | praw.models.Redditor:
        r"""Return a :class:`.RedditorList` of friends or a :class:`.Redditor` in the friends list.

        :param user: Checks to see if you are friends with the redditor. Either an
            instance of :class:`.Redditor` or a string can be given.

        :returns: A list of :class:`.Redditor`\ s, or a single :class:`.Redditor` if
            ``user`` is specified. The :class:`.Redditor` instance(s) returned also has
            friend attributes.

        :raises: An instance of :class:`.RedditAPIException` if you are not friends with
            the specified :class:`.Redditor`.

        """
        endpoint = (
            API_PATH["friends"]
            if user is None
            else API_PATH["friend_v1"].format(user=str(user))
        )
        return self._reddit.get(endpoint)

    def karma(self) -> dict[praw.models.Subreddit, dict[str, int]]:
        r"""Return a dictionary mapping :class:`.Subreddit`\ s to their karma.

        The returned dict contains subreddits as keys. Each subreddit key contains a
        sub-dict that have keys for ``comment_karma`` and ``link_karma``. The dict is
        sorted in descending karma order.

        .. note::

            Each key of the main dict is an instance of :class:`.Subreddit`. It is
            recommended to iterate over the dict in order to retrieve the values,
            preferably through :py:meth:`dict.items`.

        """
        karma_map = {}
        for row in self._reddit.get(API_PATH["karma"])["data"]:
            subreddit = Subreddit(self._reddit, row["sr"])
            del row["sr"]
            karma_map[subreddit] = row
        return karma_map

    @_deprecate_args("use_cache")
    def me(self, *, use_cache: bool = True) -> praw.models.Redditor | None:
        """Return a :class:`.Redditor` instance for the authenticated user.

        :param use_cache: When ``True``, and if this function has been previously
            called, returned the cached version (default: ``True``).

        .. note::

            If you change the :class:`.Reddit` instance's authorization, you might want
            to refresh the cached value. Prefer using separate :class:`.Reddit`
            instances, however, for distinct authorizations.

        .. deprecated:: 7.2

            In :attr:`.read_only` mode this method returns ``None``. In PRAW 8 this
            method will raise :class:`.ReadOnlyException` when called in
            :attr:`.read_only` mode. To operate in PRAW 8 mode, set the config variable
            ``praw8_raise_exception_on_me`` to ``True``.

        """
        if self._reddit.read_only:
            if not self._reddit.config.custom.get("praw8_raise_exception_on_me"):
                warn(
                    "The 'None' return value is deprecated, and will raise a"
                    " ReadOnlyException beginning with PRAW 8. See documentation for"
                    " forward compatibility options.",
                    category=DeprecationWarning,
                    stacklevel=2,
                )
                return None
            msg = "`user.me()` does not work in read_only mode"
            raise ReadOnlyException(msg)
        if "_me" not in self.__dict__ or not use_cache:
            user_data = self._reddit.get(API_PATH["me"])
            self._me = Redditor(self._reddit, _data=user_data)
        return self._me

    def moderator_subreddits(
        self, **generator_kwargs: str | int | dict[str, str]
    ) -> Iterator[praw.models.Subreddit]:
        """Return a :class:`.ListingGenerator` subreddits that the user moderates.

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        To print a list of the names of the subreddits you moderate, try:

        .. code-block:: python

            for subreddit in reddit.user.moderator_subreddits(limit=None):
                print(str(subreddit))

        .. seealso::

            :meth:`.Redditor.moderated`

        """
        return ListingGenerator(
            self._reddit, API_PATH["my_moderator"], **generator_kwargs
        )

    def multireddits(self) -> list[praw.models.Multireddit]:
        r"""Return a list of :class:`.Multireddit`\ s belonging to the user."""
        return self._reddit.get(API_PATH["my_multireddits"])

    def pin(
        self, submission: praw.models.Submission, *, num: int = None, state: bool = True
    ) -> praw.models.Submission:
        """Set the pin state of a submission on the authenticated user's profile.

        :param submission: An instance of :class:`.Submission` that will be
            pinned/unpinned.
        :param num: If specified, the slot in which the submission will be pinned into.
            If there is a submission already in the specified slot, it will be replaced.
            If ``None`` or there is not a submission in the specified slot, the first
            available slot will be used (default: ``None``). If all slots are used the
            following will occur:

            - Old Reddit:

              1. The submission in the last slot will be unpinned.
              2. The remaining pinned submissions will be shifted down a slot.
              3. The new submission will be pinned in the first slot.

            - New Reddit:

              1. The submission in the first slot will be unpinned.
              2. The remaining pinned submissions will be shifted up a slot.
              3. The new submission will be pinned in the last slot.

            .. note::

                At the time of writing (10/22/2021), there are 4 pin slots available and
                pins are in reverse order on old Reddit. If ``num`` is an invalid value,
                Reddit will ignore it and the same behavior will occur as if ``num`` is
                ``None``.

        :param state: ``True`` pins the submission, ``False`` unpins (default:
            ``True``).

        :returns: The pinned submission.

        :raises: ``prawcore.BadRequest`` when pinning a removed or deleted submission.
        :raises: ``prawcore.Forbidden`` when pinning a submission the authenticated user
            is not the author of.

        .. code-block:: python

            submission = next(reddit.user.me().submissions.new())
            reddit.user.pin(submission)

        """
        data = {
            "id": submission.fullname,
            "num": num,
            "state": state,
            "to_profile": True,
        }
        try:
            return self._reddit.post(API_PATH["sticky_submission"], data=data)
        except Conflict:
            pass

    def subreddits(
        self, **generator_kwargs: str | int | dict[str, str]
    ) -> Iterator[praw.models.Subreddit]:
        r"""Return a :class:`.ListingGenerator` of :class:`.Subreddit`\ s the user is subscribed to.

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        To print a list of the subreddits that you are subscribed to, try:

        .. code-block:: python

            for subreddit in reddit.user.subreddits(limit=None):
                print(str(subreddit))

        """
        return ListingGenerator(
            self._reddit, API_PATH["my_subreddits"], **generator_kwargs
        )

    def trusted(self) -> list[praw.models.Redditor]:
        r"""Return a :class:`.RedditorList` of trusted :class:`.Redditor`\ s.

        To display the usernames of your trusted users and the times at which you
        decided to trust them, try:

        .. code-block:: python

            trusted_users = reddit.user.trusted()
            for user in trusted_users:
                print(f"User: {user.name}, time: {user.date}")

        """
        return self._reddit.get(API_PATH["trusted"])



"""Provide poll-related classes."""

from __future__ import annotations

from typing import Any

from ...util import cachedproperty
from ..base import PRAWBase


class PollOption(PRAWBase):
    """Class to represent one option of a poll.

    If ``submission`` is a poll :class:`.Submission`, access the poll's options like so:

    .. code-block:: python

        poll_data = submission.poll_data

        # By index -- print the first option
        print(poll_data.options[0])

        # By ID -- print the option with ID "576797"
        print(poll_data.option("576797"))

    .. include:: ../../typical_attributes.rst

    ============== =================================================
    Attribute      Description
    ============== =================================================
    ``id``         ID of the poll option.
    ``text``       The text of the poll option.
    ``vote_count`` The number of votes the poll option has received.
    ============== =================================================

    """

    def __repr__(self) -> str:
        """Return an object initialization representation of the instance."""
        return f"PollOption(id={self.id!r})"

    def __str__(self) -> str:
        """Return a string version of the PollData, its text."""
        return self.text


class PollData(PRAWBase):
    """Class to represent poll data on a poll submission.

    If ``submission`` is a poll :class:`.Submission`, access the poll data like so:

    .. code-block:: python

        poll_data = submission.poll_data
        print(f"There are {poll_data.total_vote_count} votes total.")
        print("The options are:")
        for option in poll_data.options:
            print(f"{option} ({option.vote_count} votes)")
        print(f"I voted for {poll_data.user_selection}.")

    .. include:: ../../typical_attributes.rst

    ======================== =========================================================
    Attribute                Description
    ======================== =========================================================
    ``options``              A list of :class:`.PollOption` of the poll.
    ``total_vote_count``     The total number of votes cast in the poll.
    ``user_selection``       The poll option selected by the authenticated user
                             (possibly ``None``).
    ``voting_end_timestamp`` Time the poll voting closes, represented in `Unix Time`_.
    ======================== =========================================================

    .. _unix time: https://en.wikipedia.org/wiki/Unix_time

    """

    @cachedproperty
    def user_selection(self) -> PollOption | None:
        """Get the user's selection in this poll, if any.

        :returns: The user's selection as a :class:`.PollOption`, or ``None`` if there
            is no choice.

        """
        if self._user_selection is None:
            return None
        return self.option(self._user_selection)

    def __setattr__(self, attribute: str, value: Any):
        """Objectify the options attribute, and save user_selection."""
        if attribute == "options" and isinstance(value, list):
            value = [PollOption(self._reddit, option) for option in value]
        elif attribute == "user_selection":
            attribute = "_user_selection"
        super().__setattr__(attribute, value)

    def option(self, option_id: str) -> PollOption:
        """Get the option with the specified ID.

        :param option_id: The ID of a poll option, as a ``str``.

        :returns: The specified :class:`.PollOption`.

        :raises: :py:class:`KeyError` if no option exists with the specified ID.

        """
        for option in self.options:
            if option.id == option_id:
                return option

        msg = f"No poll option with ID {option_id!r}."
        raise KeyError(msg)



"""Python Reddit API Wrapper.

PRAW, an acronym for "Python Reddit API Wrapper", is a python package that allows for
simple access to Reddit's API. PRAW aims to be as easy to use as possible and is
designed to follow all of Reddit's API rules. You have to give a useragent, everything
else is handled by PRAW so you needn't worry about violating them.

More information about PRAW can be found at https://github.com/praw-dev/praw.

"""

from .const import __version__
from .reddit import Reddit



"""Provide Collections functionality."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ...const import API_PATH
from ...exceptions import ClientException
from ...util import _deprecate_args
from ...util.cache import cachedproperty
from ..base import PRAWBase
from .base import RedditBase
from .submission import Submission
from .subreddit import Subreddit

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Iterator

    import praw.models


class CollectionModeration(PRAWBase):
    """Class to support moderation actions on a :class:`.Collection`.

    Obtain an instance via:

    .. code-block:: python

        reddit.subreddit("test").collections("some_uuid").mod

    """

    def __init__(self, reddit: praw.Reddit, collection_id: str):
        """Initialize a :class:`.CollectionModeration` instance.

        :param collection_id: The ID of a :class:`.Collection`.

        """
        super().__init__(reddit, _data=None)
        self.collection_id = collection_id

    def _post_fullname(self, post: str | praw.models.Submission) -> str:
        """Get a post's fullname.

        :param post: A fullname, a :class:`.Submission`, a permalink, or an ID.

        :returns: The fullname of the post.

        """
        if isinstance(post, Submission):
            return post.fullname
        if not isinstance(post, str):
            msg = f"Cannot get fullname from object of type {type(post)}."
            raise TypeError(msg)
        if post.startswith(f"{self._reddit.config.kinds['submission']}_"):
            return post
        try:
            return self._reddit.submission(url=post).fullname
        except ClientException:
            return self._reddit.submission(post).fullname

    def add_post(self, submission: praw.models.Submission):
        """Add a post to the collection.

        :param submission: The post to add, a :class:`.Submission`, its permalink as a
            ``str``, its fullname as a ``str``, or its ID as a ``str``.

        Example usage:

        .. code-block:: python

            collection = reddit.subreddit("test").collections("some_uuid")
            collection.mod.add_post("bgibu9")

        .. seealso::

            :meth:`.remove_post`

        """
        link_fullname = self._post_fullname(submission)

        self._reddit.post(
            API_PATH["collection_add_post"],
            data={"collection_id": self.collection_id, "link_fullname": link_fullname},
        )

    def delete(self):
        """Delete this collection.

        Example usage:

        .. code-block:: python

            reddit.subreddit("test").collections("some_uuid").mod.delete()

        .. seealso::

            :meth:`~.SubredditCollectionsModeration.create`

        """
        self._reddit.post(
            API_PATH["collection_delete"], data={"collection_id": self.collection_id}
        )

    def remove_post(self, submission: praw.models.Submission):
        """Remove a post from the collection.

        :param submission: The post to remove, a :class:`.Submission`, its permalink as
            a ``str``, its fullname as a ``str``, or its ID as a ``str``.

        Example usage:

        .. code-block:: python

            collection = reddit.subreddit("test").collections("some_uuid")
            collection.mod.remove_post("bgibu9")

        .. seealso::

            :meth:`.add_post`

        """
        link_fullname = self._post_fullname(submission)

        self._reddit.post(
            API_PATH["collection_remove_post"],
            data={"collection_id": self.collection_id, "link_fullname": link_fullname},
        )

    def reorder(self, links: list[str | praw.models.Submission]):
        r"""Reorder posts in the collection.

        :param links: A list of :class:`.Submission`\ s or a ``str`` that is either a
            fullname or an ID.

        Example usage:

        .. code-block:: python

            collection = reddit.subreddit("test").collections("some_uuid")
            current_order = collection.link_ids
            new_order = reversed(current_order)
            collection.mod.reorder(new_order)

        """
        link_ids = ",".join(self._post_fullname(post) for post in links)
        self._reddit.post(
            API_PATH["collection_reorder"],
            data={"collection_id": self.collection_id, "link_ids": link_ids},
        )

    def update_description(self, description: str):
        """Update the collection's description.

        :param description: The new description.

        Example usage:

        .. code-block:: python

            collection = reddit.subreddit("test").collections("some_uuid")
            collection.mod.update_description("Please enjoy these links!")

        .. seealso::

            :meth:`.update_title`

        """
        self._reddit.post(
            API_PATH["collection_desc"],
            data={"collection_id": self.collection_id, "description": description},
        )

    def update_display_layout(self, display_layout: str):
        """Update the collection's display layout.

        :param display_layout: Either ``"TIMELINE"`` for events or discussions or
            ``"GALLERY"`` for images or memes. Passing ``None`` will clear the set
            layout and ``collection.display_layout`` will be ``None``, however, the
            collection will appear on Reddit as if ``display_layout`` is set to
            ``"TIMELINE"``.

        Example usage:

        .. code-block:: python

            collection = reddit.subreddit("test").collections("some_uuid")
            collection.mod.update_display_layout("GALLERY")

        """
        self._reddit.post(
            API_PATH["collection_layout"],
            data={
                "collection_id": self.collection_id,
                "display_layout": display_layout,
            },
        )

    def update_title(self, title: str):
        """Update the collection's title.

        :param title: The new title.

        Example usage:

        .. code-block:: python

            collection = reddit.subreddit("test").collections("some_uuid")
            collection.mod.update_title("Titley McTitleface")

        .. seealso::

            :meth:`.update_description`

        """
        self._reddit.post(
            API_PATH["collection_title"],
            data={"collection_id": self.collection_id, "title": title},
        )


class SubredditCollectionsModeration(PRAWBase):
    r"""Class to represent moderator actions on a :class:`.Subreddit`'s :class:`.Collection`\ s.

    Obtain an instance via:

    .. code-block:: python

        reddit.subreddit("test").collections.mod

    """

    def __init__(
        self,
        reddit: praw.Reddit,
        sub_fullname: str,
        _data: dict[str, Any] | None = None,
    ):
        """Initialize a :class:`.SubredditCollectionsModeration` instance."""
        super().__init__(reddit, _data)
        self.subreddit_fullname = sub_fullname

    @_deprecate_args("title", "description", "display_layout")
    def create(
        self, *, description: str, display_layout: str | None = None, title: str
    ) -> Collection:
        """Create a new :class:`.Collection`.

        The authenticated account must have appropriate moderator permissions in the
        subreddit this collection belongs to.

        :param description: The description, up to 500 characters.
        :param display_layout: Either ``"TIMELINE"`` for events or discussions or
            ``"GALLERY"`` for images or memes. Passing ``""`` or ``None`` will make the
            collection appear on Reddit as if this is set to ``"TIMELINE"`` (default:
            ``None``).
        :param title: The title of the collection, up to 300 characters.

        :returns: The newly created :class:`.Collection`.

        Example usage:

        .. code-block:: python

            my_sub = reddit.subreddit("test")
            new_collection = my_sub.collections.mod.create(title="Title", description="desc")
            new_collection.mod.add_post("bgibu9")

        To specify the display layout as ``"GALLERY"`` when creating the collection:

        .. code-block:: python

            my_sub = reddit.subreddit("test")
            new_collection = my_sub.collections.mod.create(
                title="Title", description="desc", display_layout="GALLERY"
            )
            new_collection.mod.add_post("bgibu9")

        .. seealso::

            :meth:`~.CollectionModeration.delete`

        """
        data = {
            "sr_fullname": self.subreddit_fullname,
            "title": title,
            "description": description,
        }
        if display_layout:
            data["display_layout"] = display_layout
        return self._reddit.post(
            API_PATH["collection_create"],
            data=data,
        )


class SubredditCollections(PRAWBase):
    r"""Class to represent a :class:`.Subreddit`'s :class:`.Collection`\ s.

    Obtain an instance via:

    .. code-block:: python

        reddit.subreddit("test").collections

    """

    @cachedproperty
    def mod(self) -> SubredditCollectionsModeration:
        """Get an instance of :class:`.SubredditCollectionsModeration`.

        Provides :meth:`~SubredditCollectionsModeration.create`:

        .. code-block:: python

            my_sub = reddit.subreddit("test")
            new_collection = my_sub.collections.mod.create(title="Title", description="desc")

        """
        return SubredditCollectionsModeration(self._reddit, self.subreddit.fullname)

    def __call__(
        self,
        collection_id: str | None = None,
        permalink: str | None = None,
    ) -> Collection:
        """Return the :class:`.Collection` with the specified ID.

        :param collection_id: The ID of a :class:`.Collection` (default: ``None``).
        :param permalink: The permalink of a collection (default: ``None``).

        :returns: The specified :class:`.Collection`.

        Exactly one of ``collection_id`` or ``permalink`` is required.

        Example usage:

        .. code-block:: python

            subreddit = reddit.subreddit("test")

            uuid = "847e4548-a3b5-4ad7-afb4-edbfc2ed0a6b"
            collection = subreddit.collections(uuid)
            print(collection.title)
            print(collection.description)

            permalink = "https://www.reddit.com/r/SUBREDDIT/collection/" + uuid
            collection = subreddit.collections(permalink=permalink)
            print(collection.title)
            print(collection.description)

        """
        if (collection_id is None) == (permalink is None):
            msg = "Exactly one of 'collection_id' or 'permalink' must be provided."
            raise TypeError(msg)
        return Collection(
            self._reddit, collection_id=collection_id, permalink=permalink
        )

    def __init__(
        self,
        reddit: praw.Reddit,
        subreddit: praw.models.Subreddit,
        _data: dict[str, Any] | None = None,
    ):
        """Initialize a :class:`.SubredditCollections` instance."""
        super().__init__(reddit, _data)
        self.subreddit = subreddit

    def __iter__(self):
        r"""Iterate over the :class:`.Subreddit`'s :class:`.Collection`\ s.

        Example usage:

        .. code-block:: python

            for collection in reddit.subreddit("test").collections:
                print(collection.permalink)

        """
        request = self._reddit.get(
            API_PATH["collection_subreddit"],
            params={"sr_fullname": self.subreddit.fullname},
        )
        yield from request


class Collection(RedditBase):
    """Class to represent a :class:`.Collection`.

    Obtain an instance via:

    .. code-block:: python

        collection = reddit.subreddit("test").collections("some_uuid")

    or

    .. code-block:: python

        collection = reddit.subreddit("test").collections(
            permalink="https://reddit.com/r/SUBREDDIT/collection/some_uuid"
        )

    .. include:: ../../typical_attributes.rst

    =================== =============================================================
    Attribute           Description
    =================== =============================================================
    ``author``          The :class:`.Redditor` who created the collection.
    ``collection_id``   The UUID of the collection.
    ``created_at_utc``  Time the collection was created, represented in `Unix Time`_.
    ``description``     The collection description.
    ``display_layout``  The collection display layout.
    ``last_update_utc`` Time the collection was last updated, represented in `Unix
                        Time`_.
    ``link_ids``        A list of :class:`.Submission` fullnames.
    ``permalink``       The collection's permalink (to view on the web).
    ``sorted_links``    An iterable listing of the posts in this collection.
    ``title``           The title of the collection.
    =================== =============================================================

    .. _unix time: https://en.wikipedia.org/wiki/Unix_time

    """

    STR_FIELD = "collection_id"

    @cachedproperty
    def mod(self) -> CollectionModeration:
        """Get an instance of :class:`.CollectionModeration`.

        Provides access to various methods, including
        :meth:`~.CollectionModeration.add_post`, :meth:`~.CollectionModeration.delete`,
        :meth:`~.CollectionModeration.reorder`, and
        :meth:`~.CollectionModeration.update_title`.

        Example usage:

        .. code-block:: python

            collection = reddit.subreddit("test").collections("some_uuid")
            collection.mod.update_title("My new title!")

        """
        return CollectionModeration(self._reddit, self.collection_id)

    @cachedproperty
    def subreddit(self) -> praw.models.Subreddit:
        """Get the subreddit that this collection belongs to.

        For example:

        .. code-block:: python

            collection = reddit.subreddit("test").collections("some_uuid")
            subreddit = collection.subreddit

        """
        return next(self._reddit.info(fullnames=[self.subreddit_id]))

    def __init__(
        self,
        reddit: praw.Reddit,
        _data: dict[str, Any] = None,
        collection_id: str | None = None,
        permalink: str | None = None,
    ):
        """Initialize a :class:`.Collection` instance.

        :param reddit: An instance of :class:`.Reddit`.
        :param _data: Any data associated with the :class:`.Collection`.
        :param collection_id: The ID of the :class:`.Collection`.
        :param permalink: The permalink of the :class:`.Collection`.

        """
        if (_data, collection_id, permalink).count(None) != 2:
            msg = "Exactly one of '_data', 'collection_id', or 'permalink' must be provided."
            raise TypeError(msg)

        if permalink:
            collection_id = self._url_parts(permalink)[4]

        if collection_id:
            self.collection_id = collection_id  # set from _data otherwise

        super().__init__(reddit, _data)

        self._info_params = {
            "collection_id": self.collection_id,
            "include_links": True,
        }

    def __iter__(self) -> Iterator:
        """Provide a way to iterate over the posts in this :class:`.Collection`.

        Example usage:

        .. code-block:: python

            collection = reddit.subreddit("test").collections("some_uuid")
            for submission in collection:
                print(submission.title, submission.permalink)

        """
        yield from self.sorted_links

    def __len__(self) -> int:
        """Get the number of posts in this :class:`.Collection`.

        Example usage:

        .. code-block:: python

            collection = reddit.subreddit("test").collections("some_uuid")
            print(len(collection))

        """
        return len(self.link_ids)

    def __setattr__(self, attribute: str, value: Any):
        """Objectify author, subreddit, and sorted_links attributes."""
        if attribute == "author_name":
            self.author = self._reddit.redditor(value)
        elif attribute == "sorted_links":
            value = self._reddit._objector.objectify(value)
        super().__setattr__(attribute, value)

    def _fetch(self):
        data = self._fetch_data()
        try:
            self._reddit._objector.check_error(data)
        except ClientException:
            # A well-formed but invalid Collections ID during fetch time
            # causes Reddit to return something that looks like an error
            # but with no content.
            msg = f"Error during fetch. Check collection ID {self.collection_id!r} is correct."
            raise ClientException(msg) from None

        other = type(self)(self._reddit, _data=data)
        self.__dict__.update(other.__dict__)
        super()._fetch()

    def _fetch_info(self):
        return "collection", {}, self._info_params

    def follow(self):
        """Follow this :class:`.Collection`.

        Example usage:

        .. code-block:: python

            reddit.subreddit("test").collections("some_uuid").follow()

        .. seealso::

            :meth:`.unfollow`

        """
        self._reddit.post(
            API_PATH["collection_follow"],
            data={"collection_id": self.collection_id, "follow": True},
        )

    def unfollow(self):
        """Unfollow this :class:`.Collection`.

        Example usage:

        .. code-block:: python

            reddit.subreddit("test").collections("some_uuid").unfollow()

        .. seealso::

            :meth:`.follow`

        """
        self._reddit.post(
            API_PATH["collection_follow"],
            data={"collection_id": self.collection_id, "follow": False},
        )


Subreddit._subreddit_collections_class = SubredditCollections



"""Provide the Submission class."""

from __future__ import annotations

import re
from json import dumps
from typing import TYPE_CHECKING, Any, Generator
from urllib.parse import urljoin
from warnings import warn

from prawcore import Conflict

from ...const import API_PATH
from ...exceptions import InvalidURL
from ...util import _deprecate_args, cachedproperty
from ..comment_forest import CommentForest
from ..listing.listing import Listing
from ..listing.mixins import SubmissionListingMixin
from .base import RedditBase
from .mixins import FullnameMixin, ModNoteMixin, ThingModerationMixin, UserContentMixin
from .poll import PollData
from .redditor import Redditor
from .subreddit import Subreddit

if TYPE_CHECKING:  # pragma: no cover
    import praw.models

INLINE_MEDIA_PATTERN = re.compile(
    r"\n\n!?(\[.*?])?\(?((https://((preview|i)\.redd\.it|reddit.com/link).*?)|(?!https)([a-zA-Z0-9]+( \".*?\")?))\)?"
)
MEDIA_TYPE_MAPPING = {
    "Image": "img",
    "RedditVideo": "video",
    "AnimatedImage": "gif",
}


class SubmissionFlair:
    """Provide a set of functions pertaining to :class:`.Submission` flair."""

    def __init__(self, submission: praw.models.Submission):
        """Initialize a :class:`.SubmissionFlair` instance.

        :param submission: The :class:`.Submission` associated with the flair functions.

        """
        self.submission = submission

    def choices(self) -> list[dict[str, bool | list | str]]:
        """Return list of available flair choices.

        Choices are required in order to use :meth:`.select`.

        For example:

        .. code-block:: python

            choices = submission.flair.choices()

        """
        url = API_PATH["flairselector"].format(subreddit=self.submission.subreddit)
        return self.submission._reddit.post(
            url, data={"link": self.submission.fullname}
        )["choices"]

    @_deprecate_args("flair_template_id", "text")
    def select(self, flair_template_id: str, *, text: str | None = None):
        """Select flair for submission.

        :param flair_template_id: The flair template to select. The possible values can
            be discovered through :meth:`.choices`.
        :param text: If the template's ``flair_text_editable`` value is ``True``, this
            value will set a custom text (default: ``None``).

        For example, to select an arbitrary editable flair text (assuming there is one)
        and set a custom value try:

        .. code-block:: python

            choices = submission.flair.choices()
            template_id = next(x for x in choices if x["flair_text_editable"])["flair_template_id"]
            submission.flair.select(template_id, text="my custom value")

        """
        data = {
            "flair_template_id": flair_template_id,
            "link": self.submission.fullname,
            "text": text,
        }
        url = API_PATH["select_flair"].format(subreddit=self.submission.subreddit)
        self.submission._reddit.post(url, data=data)


class SubmissionModeration(ThingModerationMixin, ModNoteMixin):
    """Provide a set of functions pertaining to :class:`.Submission` moderation.

    Example usage:

    .. code-block:: python

        submission = reddit.submission("8dmv8z")
        submission.mod.approve()

    """

    REMOVAL_MESSAGE_API = "removal_link_message"

    def __init__(self, submission: praw.models.Submission):
        """Initialize a :class:`.SubmissionModeration` instance.

        :param submission: The submission to moderate.

        """
        self.thing = submission

    @_deprecate_args("state")
    def contest_mode(self, *, state: bool = True):
        """Set contest mode for the comments of this submission.

        :param state: ``True`` enables contest mode and ``False`` disables (default:
            ``True``).

        Contest mode have the following effects:

        - The comment thread will default to being sorted randomly.
        - Replies to top-level comments will be hidden behind "[show replies]" buttons.
        - Scores will be hidden from non-moderators.
        - Scores accessed through the API (mobile apps, bots) will be obscured to "1"
          for non-moderators.

        Example usage:

        .. code-block:: python

            submission = reddit.submission("5or86n")
            submission.mod.contest_mode()

        """
        self.thing._reddit.post(
            API_PATH["contest_mode"], data={"id": self.thing.fullname, "state": state}
        )

    @_deprecate_args("text", "css_class", "flair_template_id")
    def flair(
        self,
        *,
        css_class: str = "",
        flair_template_id: str | None = None,
        text: str = "",
    ):
        """Set flair for the submission.

        :param css_class: The css class to associate with the flair html (default:
            ``""``).
        :param flair_template_id: The flair template ID to use when flairing.
        :param text: The flair text to associate with the :class:`.Submission` (default:
            ``""``).

        This method can only be used by an authenticated user who is a moderator of the
        submission's :class:`.Subreddit`.

        Example usage:

        .. code-block:: python

            submission = reddit.submission("5or86n")
            submission.mod.flair(text="PRAW", css_class="bot")

        """
        data = {
            "css_class": css_class,
            "link": self.thing.fullname,
            "text": text,
        }
        url = API_PATH["flair"].format(subreddit=self.thing.subreddit)
        if flair_template_id is not None:
            data["flair_template_id"] = flair_template_id
            url = API_PATH["select_flair"].format(subreddit=self.thing.subreddit)
        self.thing._reddit.post(url, data=data)

    def nsfw(self):
        """Mark as not safe for work.

        This method can be used both by the submission author and moderators of the
        subreddit that the submission belongs to.

        Example usage:

        .. code-block:: python

            submission = reddit.subreddit("test").submit("nsfw test", selftext="nsfw")
            submission.mod.nsfw()

        .. seealso::

            :meth:`.sfw`

        """
        self.thing._reddit.post(API_PATH["marknsfw"], data={"id": self.thing.fullname})

    def set_original_content(self):
        """Mark as original content.

        This method can be used by moderators of the subreddit that the submission
        belongs to. If the subreddit has enabled the Original Content beta feature in
        settings, then the submission's author can use it as well.

        Example usage:

        .. code-block:: python

            submission = reddit.subreddit("test").submit("oc test", selftext="original")
            submission.mod.set_original_content()

        .. seealso::

            :meth:`.unset_original_content`

        """
        data = {
            "id": self.thing.id,
            "fullname": self.thing.fullname,
            "should_set_oc": True,
            "executed": False,
            "r": self.thing.subreddit,
        }
        self.thing._reddit.post(API_PATH["set_original_content"], data=data)

    def sfw(self):
        """Mark as safe for work.

        This method can be used both by the submission author and moderators of the
        subreddit that the submission belongs to.

        Example usage:

        .. code-block:: python

            submission = reddit.submission("5or86n")
            submission.mod.sfw()

        .. seealso::

            :meth:`.nsfw`

        """
        self.thing._reddit.post(
            API_PATH["unmarknsfw"], data={"id": self.thing.fullname}
        )

    def spoiler(self):
        """Indicate that the submission contains spoilers.

        This method can be used both by the submission author and moderators of the
        subreddit that the submission belongs to.

        Example usage:

        .. code-block:: python

            submission = reddit.submission("5or86n")
            submission.mod.spoiler()

        .. seealso::

            :meth:`.unspoiler`

        """
        self.thing._reddit.post(API_PATH["spoiler"], data={"id": self.thing.fullname})

    @_deprecate_args("state", "bottom")
    def sticky(
        self, *, bottom: bool = True, state: bool = True
    ) -> praw.models.Submission:
        """Set the submission's sticky state in its subreddit.

        :param bottom: When ``True``, set the submission as the bottom sticky. If no top
            sticky exists, this submission will become the top sticky regardless
            (default: ``True``).
        :param state: ``True`` sets the sticky for the submission and ``False`` unsets
            (default: ``True``).

        :returns: The stickied submission object.

        .. note::

            When a submission is stickied two or more times, the Reddit API responds
            with a 409 error that is raised as a ``Conflict`` by prawcore. This method
            suppresses these ``Conflict`` errors.

        This submission will replace the second stickied submission if one exists.

        For example:

        .. code-block:: python

            submission = reddit.submission("5or86n")
            submission.mod.sticky()

        """
        data = {"id": self.thing.fullname, "state": state}
        if not bottom:
            data["num"] = 1
        try:
            return self.thing._reddit.post(API_PATH["sticky_submission"], data=data)
        except Conflict:
            pass

    @_deprecate_args("sort")
    def suggested_sort(self, *, sort: str = "blank"):
        """Set the suggested sort for the comments of the submission.

        :param sort: Can be one of: ``"confidence"``, ``"top"``, ``"new"``,
            ``"controversial"``, ``"old"``, ``"random"``, ``"qa"``, or ``"blank"``
            (default: ``"blank"``).

        """
        self.thing._reddit.post(
            API_PATH["suggested_sort"], data={"id": self.thing.fullname, "sort": sort}
        )

    def unset_original_content(self):
        """Indicate that the submission is not original content.

        This method can be used by moderators of the subreddit that the submission
        belongs to. If the subreddit has enabled the Original Content beta feature in
        settings, then the submission's author can use it as well.

        Example usage:

        .. code-block:: python

            submission = reddit.subreddit("test").submit("oc test", selftext="original")
            submission.mod.unset_original_content()

        .. seealso::

            :meth:`.set_original_content`

        """
        data = {
            "id": self.thing.id,
            "fullname": self.thing.fullname,
            "should_set_oc": False,
            "executed": False,
            "r": self.thing.subreddit,
        }
        self.thing._reddit.post(API_PATH["set_original_content"], data=data)

    def unspoiler(self):
        """Indicate that the submission does not contain spoilers.

        This method can be used both by the submission author and moderators of the
        subreddit that the submission belongs to.

        For example:

        .. code-block:: python

            submission = reddit.subreddit("test").submit("not spoiler", selftext="spoiler")
            submission.mod.unspoiler()

        .. seealso::

            :meth:`.spoiler`

        """
        self.thing._reddit.post(API_PATH["unspoiler"], data={"id": self.thing.fullname})

    def update_crowd_control_level(self, level: int):
        """Change the Crowd Control level of the submission.

        :param level: An integer between 0 and 3.

        **Level Descriptions**

        ===== ======== ================================================================
        Level Name     Description
        ===== ======== ================================================================
        0     Off      Crowd Control will not action any of the submission's comments.
        1     Lenient  Comments from users who have negative karma in the subreddit are
                       automatically collapsed.
        2     Moderate Comments from new users and users with negative karma in the
                       subreddit are automatically collapsed.
        3     Strict   Comments from users who haven’t joined the subreddit, new users,
                       and users with negative karma in the subreddit are automatically
                       collapsed.
        ===== ======== ================================================================

        Example usage:

        .. code-block:: python

            submission = reddit.submission("745ryj")
            submission.mod.update_crowd_control_level(2)

        .. seealso::

            :meth:`~.CommentModeration.show`

        """
        self.thing._reddit.post(
            API_PATH["update_crowd_control"],
            data={"id": self.thing.fullname, "level": level},
        )


class Submission(SubmissionListingMixin, UserContentMixin, FullnameMixin, RedditBase):
    """A class for submissions to Reddit.

    .. include:: ../../typical_attributes.rst

    ========================== =========================================================
    Attribute                  Description
    ========================== =========================================================
    ``author``                 Provides an instance of :class:`.Redditor`.
    ``author_flair_text``      The text content of the author's flair, or ``None`` if
                               not flaired.
    ``clicked``                Whether or not the submission has been clicked by the
                               client.
    ``comments``               Provides an instance of :class:`.CommentForest`.
    ``created_utc``            Time the submission was created, represented in `Unix
                               Time`_.
    ``distinguished``          Whether or not the submission is distinguished.
    ``edited``                 Whether or not the submission has been edited.
    ``id``                     ID of the submission.
    ``is_original_content``    Whether or not the submission has been set as original
                               content.
    ``is_self``                Whether or not the submission is a selfpost (text-only).
    ``link_flair_template_id`` The link flair's ID.
    ``link_flair_text``        The link flair's text content, or ``None`` if not
                               flaired.
    ``locked``                 Whether or not the submission has been locked.
    ``name``                   Fullname of the submission.
    ``num_comments``           The number of comments on the submission.
    ``over_18``                Whether or not the submission has been marked as NSFW.
    ``permalink``              A permalink for the submission.
    ``poll_data``              A :class:`.PollData` object representing the data of this
                               submission, if it is a poll submission.
    ``saved``                  Whether or not the submission is saved.
    ``score``                  The number of upvotes for the submission.
    ``selftext``               The submissions' selftext - an empty string if a link
                               post.
    ``spoiler``                Whether or not the submission has been marked as a
                               spoiler.
    ``stickied``               Whether or not the submission is stickied.
    ``subreddit``              Provides an instance of :class:`.Subreddit`.
    ``title``                  The title of the submission.
    ``upvote_ratio``           The percentage of upvotes from all votes on the
                               submission.
    ``url``                    The URL the submission links to, or the permalink if a
                               selfpost.
    ========================== =========================================================

    .. _unix time: https://en.wikipedia.org/wiki/Unix_time

    """

    STR_FIELD = "id"

    @staticmethod
    def id_from_url(url: str) -> str:
        """Return the ID contained within a submission URL.

        :param url: A url to a submission in one of the following formats (http urls
            will also work):

            - ``"https://redd.it/2gmzqe"``
            - ``"https://reddit.com/comments/2gmzqe/"``
            - ``"https://www.reddit.com/r/redditdev/comments/2gmzqe/praw_https/"``
            - ``"https://www.reddit.com/gallery/2gmzqe"``

        :raises: :class:`.InvalidURL` if ``url`` is not a valid submission URL.

        """
        parts = RedditBase._url_parts(url)
        if "comments" not in parts and "gallery" not in parts:
            submission_id = parts[-1]
            if "r" in parts:
                raise InvalidURL(
                    url, message="Invalid URL (subreddit, not submission): {}"
                )

        elif "gallery" in parts:
            submission_id = parts[parts.index("gallery") + 1]

        elif parts[-1] == "comments":
            raise InvalidURL(url, message="Invalid URL (submission ID not present): {}")

        else:
            submission_id = parts[parts.index("comments") + 1]

        if not submission_id.isalnum():
            raise InvalidURL(url)
        return submission_id

    @cachedproperty
    def flair(self) -> SubmissionFlair:
        """Provide an instance of :class:`.SubmissionFlair`.

        This attribute is used to work with flair as a regular user of the subreddit the
        submission belongs to. Moderators can directly use :meth:`.flair`.

        For example, to select an arbitrary editable flair text (assuming there is one)
        and set a custom value try:

        .. code-block:: python

            choices = submission.flair.choices()
            template_id = next(x for x in choices if x["flair_text_editable"])["flair_template_id"]
            submission.flair.select(template_id, text="my custom value")

        """
        return SubmissionFlair(self)

    @cachedproperty
    def mod(self) -> SubmissionModeration:
        """Provide an instance of :class:`.SubmissionModeration`.

        Example usage:

        .. code-block:: python

            submission = reddit.submission("8dmv8z")
            submission.mod.approve()

        """
        return SubmissionModeration(self)

    @property
    def _kind(self) -> str:
        """Return the class's kind."""
        return self._reddit.config.kinds["submission"]

    @property
    def comments(self) -> CommentForest:
        """Provide an instance of :class:`.CommentForest`.

        This attribute can be used, for example, to obtain a flat list of comments, with
        any :class:`.MoreComments` removed:

        .. code-block:: python

            submission.comments.replace_more(limit=0)
            comments = submission.comments.list()

        Sort order and comment limit can be set with the ``comment_sort`` and
        ``comment_limit`` attributes before comments are fetched, including any call to
        :meth:`.replace_more`:

        .. code-block:: python

            submission.comment_sort = "new"
            comments = submission.comments.list()

        .. note::

            The appropriate values for ``"comment_sort"`` include ``"confidence"``,
            ``"controversial"``, ``"new"``, ``"old"``, ``"q&a"``, and ``"top"``

        See :ref:`extracting_comments` for more on working with a
        :class:`.CommentForest`.

        """
        # This assumes _comments is set so that _fetch is called when it's not.
        return self._comments

    @property
    def shortlink(self) -> str:
        """Return a shortlink to the submission.

        For example, https://redd.it/eorhm is a shortlink for
        https://www.reddit.com/r/announcements/comments/eorhm/reddit_30_less_typing/.

        """
        return urljoin(self._reddit.config.short_url, self.id)

    def __init__(
        self,
        reddit: praw.Reddit,
        id: str | None = None,
        url: str | None = None,
        _data: dict[str, Any] | None = None,
    ):
        """Initialize a :class:`.Submission` instance.

        :param reddit: An instance of :class:`.Reddit`.
        :param id: A reddit base36 submission ID, e.g., ``"2gmzqe"``.
        :param url: A URL supported by :meth:`.id_from_url`.

        Either ``id`` or ``url`` can be provided, but not both.

        """
        if (id, url, _data).count(None) != 2:
            msg = "Exactly one of 'id', 'url', or '_data' must be provided."
            raise TypeError(msg)
        self.comment_limit = 2048

        # Specify the sort order for ``comments``
        self.comment_sort = "confidence"

        if id:
            self.id = id
        elif url:
            self.id = self.id_from_url(url)

        super().__init__(reddit, _data=_data)

        self._additional_fetch_params = {}
        self._comments_by_id = {}

    def __setattr__(self, attribute: str, value: Any):
        """Objectify author, subreddit, and poll data attributes."""
        if attribute == "author":
            value = Redditor.from_data(self._reddit, value)
        elif attribute == "subreddit":
            value = Subreddit(self._reddit, value)
        elif attribute == "poll_data":
            value = PollData(self._reddit, value)
        elif (
            attribute == "comment_sort"
            and hasattr(self, "_fetched")
            and self._fetched
            and hasattr(self, "_reddit")
            and self._reddit.config.warn_comment_sort
        ):
            warn(
                "The comments for this submission have already been fetched, so the"
                " updated comment_sort will not have any effect.",
                stacklevel=2,
            )
        super().__setattr__(attribute, value)

    def _chunk(
        self,
        *,
        chunk_size: int,
        other_submissions: list[praw.models.Submission] | None,
    ) -> Generator[str, None, None]:
        all_submissions = [self.fullname]
        if other_submissions:
            all_submissions += [x.fullname for x in other_submissions]

        for position in range(0, len(all_submissions), chunk_size):
            yield ",".join(all_submissions[position : position + 50])

    def _edit_experimental(
        self,
        body: str,
        *,
        preserve_inline_media: bool = False,
        inline_media: dict[str, praw.models.InlineMedia] | None = None,
    ) -> praw.models.Submission:
        """Replace the body of the object with ``body``.

        :param body: The Markdown formatted content for the updated object.
        :param preserve_inline_media: Attempt to preserve inline media in ``body``.

            .. danger::

                This method is experimental. It is reliant on undocumented API endpoints
                and may result in existing inline media not displaying correctly and/or
                creating a malformed body. Use at your own risk. This method may be
                removed in the future without warning.

        :param inline_media: A dict of :class:`.InlineMedia` objects where the key is
            the placeholder name in ``body``.

        :returns: The current instance after updating its attributes.

        Example usage:

        .. code-block:: python

            from praw.models import InlineGif, InlineImage, InlineVideo

            submission = reddit.submission("5or86n")
            gif = InlineGif(path="path/to/image.gif", caption="optional caption")
            image = InlineImage(path="path/to/image.jpg", caption="optional caption")
            video = InlineVideo(path="path/to/video.mp4", caption="optional caption")
            body = "New body with a gif {gif1} an image {image1} and a video {video1} inline"
            media = {"gif1": gif, "image1": image, "video1": video}
            submission._edit_experimental(submission.selftext + body, inline_media=media)

        """
        data = {
            "thing_id": self.fullname,
            "validate_on_submit": self._reddit.validate_on_submit,
        }
        is_richtext_json = False
        if INLINE_MEDIA_PATTERN.search(body) and self.media_metadata:
            is_richtext_json = True
        if inline_media:
            body = body.format(
                **{
                    placeholder: self.subreddit._upload_inline_media(media)
                    for placeholder, media in inline_media.items()
                }
            )
            is_richtext_json = True
        if is_richtext_json:
            richtext_json = self.subreddit._convert_to_fancypants(body)
            if preserve_inline_media:
                self._replace_richtext_links(richtext_json)
            data["richtext_json"] = dumps(richtext_json)
        else:
            data["text"] = body
        updated = self._reddit.post(API_PATH["edit"], data=data)
        if not is_richtext_json:
            updated = updated[0]
            for attribute in [
                "_fetched",
                "_reddit",
                "_submission",
                "replies",
                "subreddit",
            ]:
                if attribute in updated.__dict__:
                    delattr(updated, attribute)
            self.__dict__.update(updated.__dict__)
        else:
            self.__dict__.update(updated)
        return self

    def _fetch(self):
        data = self._fetch_data()
        submission_listing, comment_listing = data
        comment_listing = Listing(self._reddit, _data=comment_listing["data"])

        submission_data = submission_listing["data"]["children"][0]["data"]
        submission = type(self)(self._reddit, _data=submission_data)
        delattr(submission, "comment_limit")
        delattr(submission, "comment_sort")
        submission._comments = CommentForest(self)

        self.__dict__.update(submission.__dict__)
        self.comments._update(comment_listing.children)
        super()._fetch()

    def _fetch_data(self):
        name, fields, params = self._fetch_info()
        params.update(self._additional_fetch_params.copy())
        path = API_PATH[name].format(**fields)
        return self._reddit.request(method="GET", params=params, path=path)

    def _fetch_info(self):
        return (
            "submission",
            {"id": self.id},
            {"limit": self.comment_limit, "sort": self.comment_sort},
        )

    def _replace_richtext_links(self, richtext_json: dict):
        parsed_media_types = {
            media_id: MEDIA_TYPE_MAPPING[value["e"]]
            for media_id, value in self.media_metadata.items()
        }

        for index, element in enumerate(richtext_json["document"][:]):
            element_items = element.get("c")
            if isinstance(element_items, str):
                assert element.get("e") in ["gif", "img", "video"], (
                    "Unexpected richtext JSON schema. Please file a bug report with"
                    " PRAW."
                )  # make sure this is an inline element
                continue  # pragma: no cover
            for item in element.get("c"):
                if item.get("e") == "link":
                    ids = set(parsed_media_types)
                    # remove extra bits from the url
                    url = item["u"].split("https://")[1].split("?")[0]
                    # the id is in the url somewhere, so we split by '/' and '.'
                    matched_id = ids.intersection(re.split(r"[./]", url))
                    if matched_id:
                        matched_id = matched_id.pop()
                        correct_element = {
                            "e": parsed_media_types[matched_id],
                            "id": matched_id,
                        }
                        if item.get("t") != item.get("u"):  # add caption if it exists
                            correct_element["c"] = item["t"]
                        richtext_json["document"][index] = correct_element

    def add_fetch_param(self, key: str, value: str):
        """Add a parameter to be used for the next fetch.

        :param key: The key of the fetch parameter.
        :param value: The value of the fetch parameter.

        For example, to fetch a submission with the ``rtjson`` attribute populated:

        .. code-block:: python

            submission = reddit.submission("mcqjl8")
            submission.add_fetch_param("rtj", "all")
            print(submission.rtjson)

        """
        if (
            hasattr(self, "_fetched")
            and self._fetched
            and hasattr(self, "_reddit")
            and self._reddit.config.warn_additional_fetch_params
        ):
            warn(
                f"This {self.__class__.__name__.lower()} has already been fetched, so"
                " adding additional fetch parameters will not have any effect.",
                stacklevel=2,
            )
        self._additional_fetch_params[key] = value

    @_deprecate_args(
        "subreddit",
        "title",
        "send_replies",
        "flair_id",
        "flair_text",
        "nsfw",
        "spoiler",
    )
    def crosspost(
        self,
        subreddit: praw.models.Subreddit,
        *,
        flair_id: str | None = None,
        flair_text: str | None = None,
        nsfw: bool = False,
        send_replies: bool = True,
        spoiler: bool = False,
        title: str | None = None,
    ) -> praw.models.Submission:
        """Crosspost the submission to a subreddit.

        .. note::

            Be aware you have to be subscribed to the target subreddit.

        :param subreddit: Name of the subreddit or :class:`.Subreddit` object to
            crosspost into.
        :param flair_id: The flair template to select (default: ``None``).
        :param flair_text: If the template's ``flair_text_editable`` value is ``True``,
            this value will set a custom text (default: ``None``).
        :param nsfw: Whether the submission should be marked NSFW (default: ``False``).
        :param send_replies: When ``True``, messages will be sent to the created
            submission's author when comments are made to the submission (default:
            ``True``).
        :param spoiler: Whether the submission should be marked as a spoiler (default:
            ``False``).
        :param title: Title of the submission. Will use this submission's title if
            ``None`` (default: ``None``).

        :returns: A :class:`.Submission` object for the newly created submission.

        Example usage:

        .. code-block:: python

            submission = reddit.submission("5or86n")
            cross_post = submission.crosspost("learnprogramming", send_replies=False)

        .. seealso::

            :meth:`.hide`

        """
        if title is None:
            title = self.title

        data = {
            "sr": str(subreddit),
            "title": title,
            "sendreplies": bool(send_replies),
            "kind": "crosspost",
            "crosspost_fullname": self.fullname,
            "nsfw": bool(nsfw),
            "spoiler": bool(spoiler),
        }
        for key, value in (("flair_id", flair_id), ("flair_text", flair_text)):
            if value is not None:
                data[key] = value

        return self._reddit.post(API_PATH["submit"], data=data)

    @_deprecate_args("other_submissions")
    def hide(self, *, other_submissions: list[praw.models.Submission] | None = None):
        """Hide :class:`.Submission`.

        :param other_submissions: When provided, additionally hide this list of
            :class:`.Submission` instances as part of a single request (default:
            ``None``).

        Example usage:

        .. code-block:: python

            submission = reddit.submission("5or86n")
            submission.hide()

        .. seealso::

            :meth:`.unhide`

        """
        for submissions in self._chunk(
            chunk_size=50, other_submissions=other_submissions
        ):
            self._reddit.post(API_PATH["hide"], data={"id": submissions})

    def mark_visited(self):
        """Mark submission as visited.

        This method requires a subscription to reddit premium.

        Example usage:

        .. code-block:: python

            submission = reddit.submission("5or86n")
            submission.mark_visited()

        """
        data = {"links": self.fullname}
        self._reddit.post(API_PATH["store_visits"], data=data)

    @_deprecate_args("other_submissions")
    def unhide(self, *, other_submissions: list[praw.models.Submission] | None = None):
        """Unhide :class:`.Submission`.

        :param other_submissions: When provided, additionally unhide this list of
            :class:`.Submission` instances as part of a single request (default:
            ``None``).

        Example usage:

        .. code-block:: python

            submission = reddit.submission("5or86n")
            submission.unhide()

        .. seealso::

            :meth:`.hide`

        """
        for submissions in self._chunk(
            chunk_size=50, other_submissions=other_submissions
        ):
            self._reddit.post(API_PATH["unhide"], data={"id": submissions})


Subreddit._submission_class = Submission



"""Provide the BaseListingMixin class."""

from __future__ import annotations

from typing import Any, Iterator
from urllib.parse import urljoin

from ....util import _deprecate_args
from ...base import PRAWBase
from ..generator import ListingGenerator


class BaseListingMixin(PRAWBase):
    """Adds minimum set of methods that apply to all listing objects."""

    VALID_TIME_FILTERS = {"all", "day", "hour", "month", "week", "year"}

    @staticmethod
    def _validate_time_filter(time_filter: str):
        """Validate ``time_filter``.

        :raises: :py:class:`ValueError` if ``time_filter`` is not valid.

        """
        if time_filter not in BaseListingMixin.VALID_TIME_FILTERS:
            valid_time_filters = ", ".join(
                map("{!r}".format, BaseListingMixin.VALID_TIME_FILTERS)
            )
            msg = f"'time_filter' must be one of: {valid_time_filters}"
            raise ValueError(msg)

    def _prepare(self, *, arguments: dict[str, Any], sort: str) -> str:
        """Fix for :class:`.Redditor` methods that use a query param rather than subpath."""
        if self.__dict__.get("_listing_use_sort"):
            self._safely_add_arguments(arguments=arguments, key="params", sort=sort)
            return self._path
        return urljoin(self._path, sort)

    @_deprecate_args("time_filter")
    def controversial(
        self,
        *,
        time_filter: str = "all",
        **generator_kwargs: str | int | dict[str, str],
    ) -> Iterator[Any]:
        """Return a :class:`.ListingGenerator` for controversial items.

        :param time_filter: Can be one of: ``"all"``, ``"day"``, ``"hour"``,
            ``"month"``, ``"week"``, or ``"year"`` (default: ``"all"``).

        :raises: :py:class:`ValueError` if ``time_filter`` is invalid.

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        This method can be used like:

        .. code-block:: python

            reddit.domain("imgur.com").controversial(time_filter="week")
            reddit.multireddit(redditor="samuraisam", name="programming").controversial(
                time_filter="day"
            )
            reddit.redditor("spez").controversial(time_filter="month")
            reddit.redditor("spez").comments.controversial(time_filter="year")
            reddit.redditor("spez").submissions.controversial(time_filter="all")
            reddit.subreddit("all").controversial(time_filter="hour")

        """
        self._validate_time_filter(time_filter)
        self._safely_add_arguments(
            arguments=generator_kwargs, key="params", t=time_filter
        )
        url = self._prepare(arguments=generator_kwargs, sort="controversial")
        return ListingGenerator(self._reddit, url, **generator_kwargs)

    def hot(self, **generator_kwargs: str | int | dict[str, str]) -> Iterator[Any]:
        """Return a :class:`.ListingGenerator` for hot items.

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        This method can be used like:

        .. code-block:: python

            reddit.domain("imgur.com").hot()
            reddit.multireddit(redditor="samuraisam", name="programming").hot()
            reddit.redditor("spez").hot()
            reddit.redditor("spez").comments.hot()
            reddit.redditor("spez").submissions.hot()
            reddit.subreddit("all").hot()

        """
        generator_kwargs.setdefault("params", {})
        url = self._prepare(arguments=generator_kwargs, sort="hot")
        return ListingGenerator(self._reddit, url, **generator_kwargs)

    def new(self, **generator_kwargs: str | int | dict[str, str]) -> Iterator[Any]:
        """Return a :class:`.ListingGenerator` for new items.

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        This method can be used like:

        .. code-block:: python

            reddit.domain("imgur.com").new()
            reddit.multireddit(redditor="samuraisam", name="programming").new()
            reddit.redditor("spez").new()
            reddit.redditor("spez").comments.new()
            reddit.redditor("spez").submissions.new()
            reddit.subreddit("all").new()

        """
        generator_kwargs.setdefault("params", {})
        url = self._prepare(arguments=generator_kwargs, sort="new")
        return ListingGenerator(self._reddit, url, **generator_kwargs)

    @_deprecate_args("time_filter")
    def top(
        self,
        *,
        time_filter: str = "all",
        **generator_kwargs: str | int | dict[str, str],
    ) -> Iterator[Any]:
        """Return a :class:`.ListingGenerator` for top items.

        :param time_filter: Can be one of: ``"all"``, ``"day"``, ``"hour"``,
            ``"month"``, ``"week"``, or ``"year"`` (default: ``"all"``).

        :raises: :py:class:`ValueError` if ``time_filter`` is invalid.

        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        This method can be used like:

        .. code-block:: python

            reddit.domain("imgur.com").top(time_filter="week")
            reddit.multireddit(redditor="samuraisam", name="programming").top(time_filter="day")
            reddit.redditor("spez").top(time_filter="month")
            reddit.redditor("spez").comments.top(time_filter="year")
            reddit.redditor("spez").submissions.top(time_filter="all")
            reddit.subreddit("all").top(time_filter="hour")

        """
        self._validate_time_filter(time_filter)
        self._safely_add_arguments(
            arguments=generator_kwargs, key="params", t=time_filter
        )
        url = self._prepare(arguments=generator_kwargs, sort="top")
        return ListingGenerator(self._reddit, url, **generator_kwargs)
